unit Crypt;

interface

const
 CALG_RC4         = ((3 shl 13) or (4 shl 9) or 1);
 CALG_RC2         = ((3 shl 13) or (3 shl 9) or 2);
 def              = '  ';

function EncryptString(s: Ansistring; pass: Ansistring = def; alg: Cardinal = CALG_RC4): Ansistring;
function DecryptString(s: Ansistring; pass: Ansistring = def; alg: Cardinal = CALG_RC4): Ansistring;

implementation

const
 ADVAPI32            = 'advapi32.dll';
 PROV_RSA_FULL       = 1;
 CRYPT_VERIFYCONTEXT = $F0000000;
 CALG_SHA            = ((4 shl 13) or 0 or 4);

type
 HCRYPTPROV  = Cardinal;
 HCRYPTKEY   = Cardinal;
 ALG_ID      = Cardinal;
 PHCRYPTPROV = ^HCRYPTPROV;
 PHCRYPTKEY  = ^HCRYPTKEY;
 LPAWSTR     = PAnsiChar;
 HCRYPTHASH  = Cardinal;
 PHCRYPTHASH = ^HCRYPTHASH;

function CryptReleaseContext(hProv:HCRYPTPROV;dwFlags:LongWord):LongBool;stdcall;external ADVAPI32 name 'CryptReleaseContext';
function CryptAcquireContext(Prov:PHCRYPTPROV;Container:LPAWSTR;Provider:LPAWSTR;ProvType:LongWord;Flags:LongWord):LongBool;stdcall;external ADVAPI32 name 'CryptAcquireContextW';
function CryptEncrypt(Key:HCRYPTKEY;Hash:HCRYPTHASH;Final:LongBool;Flags:LongWord;Data:PBYTE;Len:PLongWord;BufLen:LongWord):LongBool;stdcall;external ADVAPI32 name 'CryptEncrypt';
function CryptDecrypt(Key:HCRYPTKEY;Hash:HCRYPTHASH;Final:LongBool;Flags:LongWord;Data:PBYTE;Len:PLongWord):LongBool;stdcall;external ADVAPI32 name 'CryptDecrypt';
function CryptCreateHash(Prov:HCRYPTPROV;Algid:ALG_ID;Key:HCRYPTKEY;Flags:LongInt;Hash:PHCRYPTHASH):LongBool;stdcall;external ADVAPI32 name 'CryptCreateHash';
function CryptHashData(Hash:HCRYPTHASH;Data:PBYTE;DataLen :LongInt;Flags:LongInt):LongBool;stdcall;external ADVAPI32 name 'CryptHashData';
function CryptDeriveKey(Prov:HCRYPTPROV;Algid:ALG_ID;BaseData:HCRYPTHASH;Flags:LongInt;Key:PHCRYPTKEY) :LongBool;stdcall;external ADVAPI32 name 'CryptDeriveKey';
function CryptDestroyHash(hHash :HCRYPTHASH) :LongBool;stdcall;external ADVAPI32 name 'CryptDestroyHash';

function ByteToHex(b: byte): Ansistring;
 function GetChar(b: byte): Ansichar;
 begin
   if b < 10 then
     Result := Ansichar(Ord('0') + b)
   else
     Result := Ansichar(Ord('A') - 10 + b);
 end;
begin
  Result := GetChar(b div 16) + GetChar(b mod 16);
end;

function StringToHex(const s: Ansistring): Ansistring;
var
  i: integer;
begin
  result := '';
  for i := 1 to Length(s) do
    result := result + ByteToHex(ord(s[i]));
end;

function StrToIntDef(s: Ansistring; def: integer): integer;
var
  i, c: integer;
begin
  Val(s, i, c);
  if c = 0 then
    Result := i
  else
    Result := def;
end;

function HexToString(const s: Ansistring): Ansistring;
var
 i: integer;
begin
 result := '';
 for i := 1 to Length(s) div 2 do
   try
     result := result + Ansichar(StrToIntDef('$' + copy(s, i*2-1, 2), 32));
   except
     result := result + '?';
   end;
end;

procedure InitPass(pass: Ansistring; alg: LongWord; var hProv: HCRYPTPROV; var hSKey: HCRYPTKEY);
var
 hash:  HCRYPTHASH;
begin
 CryptAcquireContext(@hProv, nil, nil, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT);
 CryptCreateHash(hProv, CALG_SHA, 0, 0, @hash);
 CryptHashData(hash, @pass[1], length(pass), 0);
 CryptDeriveKey(hProv, alg, hash, 0, @hSKey);
 CryptDestroyHash(hash);
end;

function EncryptString(s: Ansistring; pass: Ansistring = def; alg: Cardinal = CALG_RC4): Ansistring;
var
 p:  PByte;
 sz: LongWord;
 hProv: HCRYPTPROV;
 hSKey: HCRYPTKEY;
begin
 InitPass(pass, alg, hProv, hSKey);
 Insert(chr(Random(256)), s, 1);
 sz := Length(s);
 GetMem(p, sz + 8);
 move(s[1], p^, sz);
 if CryptEncrypt(hSKey, 0, true, 0, p, @sz, sz + 8) then
 begin
   SetLength(result, sz);
   move(p^, result[1], sz);
   result := StringToHex(result);
 end else result := s;
 FreeMem(p);
 CryptReleaseContext(hProv, 0);
end;

function DecryptString(s: Ansistring; pass: Ansistring = def; alg: Cardinal = CALG_RC4): Ansistring;
var
 p:  PByte;
 sz: LongWord;
 hProv: HCRYPTPROV;
 hSKey: HCRYPTKEY;
begin
 InitPass(pass, alg, hProv, hSKey);
 s := HexToString(s);
 sz := Length(s);
 GetMem(p, sz);
 move(s[1], p^, sz);
 if CryptDecrypt(hSKey, 0, true, 0, p, @sz) then
 begin
   SetLength(result, sz);
   move(p^, result[1], sz);
   delete(result, 1, 1);
 end else result := s;
 FreeMem(p);
 CryptReleaseContext(hProv, 0);
end;

initialization
 Randomize;

end.
